unit main;

{

Full Text TXT Search In Delphi - Sample Working Project
http://zarko-gajic.iz.hr/full-text-txt-search-in-delphi-firedac-sqlite-sample-working-project/

Requires:
Delphi XE and above
FireDac (http://www.embarcadero.com/products/rad-studio/firedac)
VirtualTreeView (http://www.delphi-gems.com/index.php/controls/virtual-treeview)

}


interface

uses
  FileCtrl, System.Types, System.IOUtils, IdHashMessageDigest, idHash, Generics.Collections, Generics.Defaults,

  Winapi.Windows, Winapi.Messages, System.SysUtils, System.Variants, System.Classes, Vcl.Graphics,
  Vcl.Controls, Vcl.Forms, Vcl.Dialogs, Vcl.StdCtrls, VirtualTrees,
  Vcl.ExtCtrls, Vcl.ComCtrls, Vcl.AppEvnts;

type
  TSearchResult = class
    DocumentId : integer;
    FileName : string;

    constructor Create(const documentId : integer; const fileName : string);
  end;
  TSearchResults = TObjectList<TSearchResult>;

  TTreeData = record
    Name: string;
    RelativeName : string;
    IsFolder : boolean;
  end;
  PTreeData = ^TTreeData;

  TMainForm = class(TForm)
    pnlTop: TPanel;
    pnlData: TPanel;
    pnlTree: TPanel;
    Splitter1: TSplitter;
    tree: TVirtualStringTree;
    edRootDirectory: TEdit;
    btnProcessFolder: TButton;
    memoContent: TMemo;
    pnlSearchTop: TPanel;
    ledFTSearch: TLabeledEdit;
    btnSearch: TButton;
    lvSearchResults: TListView;
    onlTreeActions: TPanel;
    btnUpdateItemInDatabase: TButton;
    pnlTreeTop: TPanel;
    lblTreeInfo: TLabel;
    procedure btnProcessFolderClick(Sender: TObject);
    procedure treeFreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure FormCreate(Sender: TObject);
    procedure treeGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
    procedure treeGetImageIndex(Sender: TBaseVirtualTree; Node: PVirtualNode; Kind: TVTImageKind; Column: TColumnIndex; var Ghosted: Boolean; var ImageIndex: Integer);
    procedure treePaintText(Sender: TBaseVirtualTree; const TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType);
    procedure treeChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
    procedure FormDestroy(Sender: TObject);
    procedure btnSearchClick(Sender: TObject);
    procedure lvSearchResultsSelectItem(Sender: TObject; Item: TListItem; Selected: Boolean);
    procedure btnUpdateItemInDatabaseClick(Sender: TObject);
  private
    fSearchResults : TSearchResults;
    fRootDirectory: string;
    procedure SetRootDirectory(const Value: string);
  private
    procedure FillTree;
    function GetSearchResults: TSearchResults;

    procedure FileLocateIterator(Sender: TBaseVirtualTree; Node: PVirtualNode; Data: Pointer; var Abort: Boolean);
  public
    procedure LogInMemo(const s : string);

    property RootDirectory : string read fRootDirectory write SetRootDirectory;
    property SearchResults : TSearchResults read GetSearchResults;
  end;

var
  MainForm: TMainForm;

  SR : TSearchResult;

implementation
{$R *.dfm}
uses uDM;

procedure TMainForm.btnProcessFolderClick(Sender: TObject);
var
  selectedRoot : string;
begin
  if NOT SelectDirectory('Select root directory', '', selectedRoot) then Exit;

  RootDirectory := selectedRoot;
end;


procedure TMainForm.btnSearchClick(Sender: TObject);
var
  sr : TSearchResult;
  slInFiles : TStringList;
  vNode : PVirtualNode;
  treeData : PTreeData;
  li : TListItem;
begin
  slInFiles := TStringList.Create;
  try
    vNode := tree.RootNode.FirstChild;
    while Assigned(vNode) do
    begin
      if vNode.CheckState = csCheckedNormal then
      begin
        treeData := PTreeData(tree.GetNodeData(vNode));
        slInFiles.Add(treeData.Name);
      end;

      vNode := tree.GetNextNoInit(vNode);
    end;

    DM.RetrieveSearchResults(ledFTSearch.Text, slInFiles);
  finally
    slInFiles.Free;
  end;

  lvSearchResults.Items.BeginUpdate;
  try
    lvSearchResults.Clear;
    for sr in SearchResults do
    begin
      li := lvSearchResults.Items.Add;
      li.Caption := IntToStr(sr.DocumentId);
      li.SubItems.Add(sr.FileName);
      li.Data := sr;
    end;
  finally
    lvSearchResults.Items.EndUpdate;
  end;
end;


procedure TMainForm.btnUpdateItemInDatabaseClick(Sender: TObject);
var
  data: PTreeData;
begin
  data := tree.GetNodeData(tree.FocusedNode);
  if (NOT Assigned(data)) OR (data.IsFolder) then
  begin
    Beep;
    Exit;
  end;

  dm.UpdateFileInDB(data.Name, true);
end;

procedure TMainForm.FileLocateIterator(Sender: TBaseVirtualTree; Node: PVirtualNode; Data: Pointer; var Abort: Boolean);
var
  treeData : PTreeData;
begin
  treeData := PTreeData(Sender.GetNodeData(Node));
  if treeData.Name = String(Data) then Abort := true;
end;

procedure TMainForm.FillTree;
  procedure ProcessDirectory(const dir : string; const parentNode : PVirtualNode);
  var
    dNode, fNode : PVirtualNode;
    subFolders, subFiles : TStringDynArray;
    subFolder, subFile : string;
    treeData : PTreeData;
  begin
    subFiles := TDirectory.GetFiles(
                    dir,
                    TSearchOption.soTopDirectoryOnly,
                    function(const Path: string; const SearchRec: TSearchRec): Boolean
                    begin
                      result := SameText(ExtractFileExt(SearchRec.Name), '.txt');
                    end);

    for subFile in subFiles do
    begin
      fNode := tree.AddChild(parentNode);
      fNode.CheckType := ctCheckBox;
      fNode.CheckState := csCheckedNormal;

      treeData := PTreeData(tree.GetNodeData(fNode));
      treeData.Name := subFile;
      treeData.RelativeName := Copy(subFile, 1 + Length(RootDirectory));
      treeData.IsFolder := false;
    end;

    subFolders := TDirectory.GetDirectories(dir, TSearchOption.soTopDirectoryOnly, nil);

    LogInMemo('> ' + dir);

    for subFolder in subFolders do
    begin
      dNode := tree.AddChild(parentNode);
      dNode.CheckType := ctTriStateCheckBox;
      dNode.CheckState := csCheckedNormal;

      treeData := PTreeData(tree.GetNodeData(dNode));
      treeData.Name := subFolder;
      treeData.RelativeName := Copy(subFolder, 1 + Length(RootDirectory));
      treeData.IsFolder := true;

      ProcessDirectory(subFolder, dNode);
    end;
  end; (*ProcessDirectory*)
begin

  tree.Clear;

  tree.BeginUpdate;
  try
    ProcessDirectory(RootDirectory, tree.RootNode);
    tree.ValidateNode(nil, True);
    tree.FullExpand();
  finally
    tree.EndUpdate;
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  tree.NodeDataSize := SizeOf(TTreeData);

  {$IFDEF DEBUG}
  RootDirectory := 'E:\ftsTXTTest\A';
  {$ENDIF}
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  fSearchResults.Free;
end;

function TMainForm.GetSearchResults: TSearchResults;
begin
  if fSearchResults = nil then fSearchResults := TSearchResults.Create(true);
  result := fSearchResults;
end;

procedure TMainForm.LogInMemo(const s: string);
begin
  memoContent.Lines.Add(Format('%d. %s', [1 + memoContent.Lines.Count ,s]));
end;

procedure TMainForm.lvSearchResultsSelectItem(Sender: TObject;
  Item: TListItem; Selected: Boolean);
var
  sr : TSearchResult;
  tNode : PVirtualNode;
begin
  if NOT Selected then Exit;

  sr := TSearchResult(Item.Data);

  tNode := tree.IterateSubtree(nil, FileLocateIterator, PChar(sr.FileName));
  if Assigned(tNode) then
  begin
    tree.ScrollIntoView(tNode, true);
    tree.Selected[tNode] := true;
    tree.FocusedNode := tNode;
  end;
end;

procedure TMainForm.SetRootDirectory(const Value: string);
begin
  fRootDirectory := Value;

  edRootDirectory.Text := RootDirectory;

  lvSearchResults.Clear;

  FillTree;

  DM.DatabaseFileName := IncludeTrailingPathDelimiter(RootDirectory) + 'fts_txt.sdb';
  DM.ConnectCreateDatabase;
end;

procedure TMainForm.treeChange(Sender: TBaseVirtualTree; Node: PVirtualNode);
var
  data: PTreeData;
begin
  data := Sender.GetNodeData(sender.FocusedNode);
  if NOT Assigned(data) then Exit;

  btnUpdateItemInDatabase.Enabled := NOT data.IsFolder;

  if data.IsFolder then
    memoContent.Clear
  else
    memoContent.Lines.LoadFromFile(data.Name);
end;

procedure TMainForm.treeFreeNode(Sender: TBaseVirtualTree; Node: PVirtualNode);
var
  data: PTreeData;
begin
  data := Sender.GetNodeData(Node);
  Finalize(data^);
end;

procedure TMainForm.treeGetImageIndex(Sender: TBaseVirtualTree; Node: PVirtualNode; Kind: TVTImageKind; Column: TColumnIndex; var Ghosted: Boolean; var ImageIndex: Integer);
var
  data: PTreeData;
begin
  data := Sender.GetNodeData(Node);

  if (Column = 0) AND (Kind in [ikNormal,ikSelected]) then
  begin
    if data.IsFolder then
      ImageIndex := 0
    else
      ImageIndex := 1;
  end;
end;

procedure TMainForm.treeGetText(Sender: TBaseVirtualTree; Node: PVirtualNode; Column: TColumnIndex; TextType: TVSTTextType; var CellText: string);
var
  Data: PTreeData;
begin
  Data := Sender.GetNodeData(Node);

  if data.IsFolder then
    CellText := data.RelativeName
  else
    CellText := ExtractFileName(data.Name);
end;

procedure TMainForm.treePaintText(Sender: TBaseVirtualTree;
  const TargetCanvas: TCanvas; Node: PVirtualNode; Column: TColumnIndex;
  TextType: TVSTTextType);
var
  data: PTreeData;
begin
  data := Sender.GetNodeData(Node);

  if (Column = 0) then
  begin
    if data.IsFolder then TargetCanvas.Font.Style := [fsBold];
  end;
end;

{ TSearchResult }

constructor TSearchResult.Create(const documentId: integer; const fileName: string);
begin
  self.DocumentId := documentId;
  self.FileName := fileName;
end;

end.
