unit uDM;

{

Full Text TXT Search In Delphi - Sample Working Project
http://zarko-gajic.iz.hr/full-text-txt-search-in-delphi-firedac-sqlite-sample-working-project/

Requires:
Delphi XE and above
FireDac (http://www.embarcadero.com/products/rad-studio/firedac)
VirtualTreeView (http://www.delphi-gems.com/index.php/controls/virtual-treeview)

}


interface

uses
  System.IOUtils, System.Types, Vcl.Dialogs, WinApi.Windows,

  System.SysUtils, System.Classes, Vcl.ImgList, Vcl.Controls, uADStanIntf,
  uADStanOption, uADStanError, uADGUIxIntf, uADPhysIntf, uADStanDef,
  uADStanPool, uADStanAsync, uADPhysManager, uADStanExprFuncs,
  uADGUIxFormsWait, uADCompGUIx, uADPhysSQLite, Data.DB, uADCompClient,
  uADStanParam, uADDatSManager, uADDAptIntf, uADDAptManager, uADCompDataSet;

type
  TDM = class(TDataModule)
    treeImages: TImageList;
    sqliteConnection: TADConnection;
    ADPhysSQLiteDriverLink1: TADPhysSQLiteDriverLink;
    ADGUIxWaitCursor1: TADGUIxWaitCursor;
    sqlQuery: TADQuery;
  private
    fDatabaseFileName: string;
  public
    property DatabaseFileName : string read fDatabaseFileName write fDatabaseFileName;
    function ConnectCreateDatabase : boolean;
    function ImportFilesToDB : boolean;
    function UpdateFileInDB(const fileName : string; const removeOld : boolean = false) : boolean;

    procedure RetrieveSearchResults(const searchToken : string; const inFiles : TStringList = nil);
  end;

var
  DM : TDM;

implementation
{$R *.dfm}

uses Main;

function TDM.ConnectCreateDatabase() : boolean;
var
  dbExists : boolean;
  r : integer;

  function DDLSQLFromResource : string;
  var
    rStream : TResourceStream;
    sStream : TStringStream;
  begin
    result := '';
    if 0 <> FindResource(hInstance, PChar('DDLSQL'), RT_RCDATA) then
    begin
      rStream := TResourceStream.Create(hInstance, 'DDLSQL', RT_RCDATA);
      sStream := TStringStream.Create;
      try
        sStream.LoadFromStream(rStream);
        result := sStream.DataString;
      finally
        rStream.Free;
        sStream.Free;
      end;
    end;
  end; (*DDLSQLFromResource*)

begin
  result := true;

  dbExists := TFile.Exists(DatabaseFileName);

  sqliteConnection.Params.Clear;
  sqliteConnection.Params.Add('DriverID=SQLite');
  sqliteConnection.Params.Add('Database=' + DatabaseFileName);
  sqliteConnection.Params.Add('OpenMode=CreateUTF16');
  sqliteConnection.Params.Add('Password=aes-256:12345');
  sqliteConnection.Params.Add('BusyTimeout=10000');
  sqliteConnection.Params.Add('LockingMode=Normal');
  sqliteConnection.Params.Add('JournalMode=WAL');
  sqliteConnection.Params.Add('Synchronous=Full');
  sqliteConnection.Params.Add('SQLiteAdvanced=temp_store=MEMORY');
  sqliteConnection.Params.Add('SQLiteAdvanced=page_size=4096');
  sqliteConnection.Params.Add('SQLiteAdvanced=auto_vacuum=FULL');

  try
    sqliteConnection.Connected := true; //will create if does not exist

    if NOT dbExists then
    begin
      r := sqliteConnection.ExecSQL(DDLSQLFromResource);

      ImportFilesToDB;
    end;
  except on E: Exception do
    begin
      result := false;
      MessageDlg('Error connecting to the database:' + E.Message, mtError, [mbOk], -1);
    end;
  end;
end;

function TDM.ImportFilesToDB: boolean;
var
  subFiles : TStringDynArray;
  subFile : string;
begin
  subFiles := TDirectory.GetFiles(
                    main.MainForm.RootDirectory,
                    TSearchOption.soAllDirectories,
                    function(const Path: string; const SearchRec: TSearchRec): Boolean
                    begin
                      result := SameText(ExtractFileExt(SearchRec.Name), '.txt');
                    end);

  for subFile in subFiles do
    UpdateFileInDB(subFile);
end;

procedure TDM.RetrieveSearchResults(const searchToken: string; const inFiles : TStringList = nil);
var
  fn : string;
  st : integer;
begin
  //get search results
  sqlQuery.SQL.Clear;

  //temp table for selected files
  sqlQuery.SQL.Add('CREATE TEMPORARY TABLE IF NOT EXISTS PartOfDocument (FileName STRING);');
  sqlQuery.SQL.Add('DELETE FROM PartOfDocument;');

  if Assigned(inFiles) then
    for fn in inFiles do
      sqlQuery.SQL.Add('INSERT INTO PartOfDocument VALUES (' + QuotedStr(fn) + ');');


  //main fts query
  sqlQuery.SQL.Add('SELECT Document.Id, Document.FileName FROM PageContent');
  sqlQuery.SQL.Add('INNER JOIN FilePages ON FilePages.rowid = PageContent.rowid');
  sqlQuery.SQL.Add('INNER JOIN Document ON Document.rowid = FilePages.DocumentID');
  if Assigned(inFiles) AND (inFiles.Count > 0) then
    sqlQuery.SQL.Add('INNER JOIN PartOfDocument ON PartOfDocument.FileName = Document.FileName');
  sqlQuery.SQL.Add('WHERE PageContent MATCH ' + QuotedStr(searchToken));
  sqlQuery.SQL.Add(';');

  sqlQuery.OpenOrExecute;

  MainForm.SearchResults.Clear;
  while NOT sqlQuery.Eof do
  begin
    MainForm.SearchResults.Add(
      TSearchResult.Create(
        sqlQuery.Fields[0].AsInteger, //doc.id
        sqlQuery.Fields[1].AsString //file name
        ));
    sqlQuery.Next;
  end;

  sqlQuery.Close;
end;

function TDM.UpdateFileInDB(const fileName: string; const removeOld : boolean = false): boolean;
var
  sStream: TStringStream;
  documentRowID : integer;
begin
  result := true;

  if removeOld then
    sqliteConnection.ExecSQL('DELETE FROM Document WHERE FileName = ' + QuotedStr(fileName));

  sStream := TStringStream.Create;
  try
    sStream.LoadFromFile(fileName);

    try
      MainForm.LogInMemo(' +' + fileName);

      sqliteConnection.ExecSQL('INSERT INTO Document VALUES (NULL, :fileName)', [fileName]);

      documentRowID := sqliteConnection.ExecSQLScalar('SELECT LAST_INSERT_ROWID() AS rowid');

      //1 page per document for the sake of simplicity
      sqliteConnection.ExecSQL('INSERT INTO FTSData VALUES (NULL, :id, :page, :txt)', [documentRowID, 1, sStream.DataString]);
    except on E: Exception do
      begin
        result := false;
        MessageDlg('Error writing to database:' + E.Message, mtError, [mbOk], -1);
      end;
    end;
  finally
    sStream.Free;
  end;
end;

end.
